from typing import Dict, List, Any, Callable, Optional
from pydantic import Field, BaseModel
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.support.ui import WebDriverWait, Select
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from webdriver_manager.chrome import ChromeDriverManager
from bs4 import BeautifulSoup
import json
import time
import re
import logging
import os
import difflib
from datetime import datetime

# Assuming base_tool is part of the project structure
# Replace with actual import path
try:
    from .base_tool import BaseTool
except ImportError:
    from base_tool import BaseTool

logger = logging.getLogger(__name__)

class BrowserPlan(BaseModel):
    tasks: List[Dict[str, Any]] = Field(..., description="List of automation tasks to execute")

def get_system_prompt(user_input: str,
                      current_url: str,
                      current_date: str,
                      interactive_elements: str = "None",
                      tabs: str = "None") -> str:
    """
    Generate a dynamic system prompt that instructs the LLM to produce a fully structured JSON plan.
    """
    prompt = f'''
You are a dynamic web automation agent that generates a fully structured JSON plan based on user input. Your response must adhere strictly to the provided JSON format and include detailed action steps to achieve the task.

**JSON Format for Response:**
{{
  "current_state": {{
    "evaluation_previous_goal": "Success | Failed | Unknown",
    "memory": "Summary of completed actions and key points.",
    "next_goal": "Description of the next steps required."
  }},
  "action": [
    {{
      "action": "navigate",
      "value": "<url id="cun213qhra013vnvh40g" type="url" status="parsed" title="Example Domain" wc="173">https://example.com</url> "
    }},
    {{
      "action": "label_ui"
    }},
    {{
      "action": "click",
      "selector": "#login-button",
      "ui_index": 2
    }},
    {{
      "done": {{}}
    }}
  ]
}}

**Available Actions:**
- navigate, label_ui, click, type, wait, wait_for_ajax, scroll, hover, screenshot, switch_tab, execute_script, drag_and_drop, select.

**Current Context:**
- Date/Time: {current_date}
- Current URL: {current_url}
- Interactive Elements: {interactive_elements}
- Open Tabs: {tabs}

**Task:**
{user_input}

Please provide your response strictly in the defined JSON format to ensure accurate execution.
'''
    return prompt.strip()

class WebBrowserTool(BaseTool):
    name: str = Field("WebBrowser", description="Tool for dynamic web automation using AI-generated plans")
    description: str = Field(
        "Generates AI-driven automation plans and executes them with advanced error recovery and context-aware interactions."
    )
    
    default_timeout: int = 30         # Increased timeout
    max_retries: int = 5              # Enhanced retry logic
    max_overall_attempts: int = 5     # Overall attempts

    synonyms_cache: Dict[str, List[str]] = {}
    ui_mapping: List[Dict[str, Any]] = []

    def execute(self, input: Dict[str, Any]) -> Dict[str, Any]:
        """
        Execute advanced dynamic web automation workflow with enhanced error handling and plan generation.
        """
        driver = None
        overall_start = time.time()
        try:
            headless = input.get("headless", False)
            self.default_timeout = int(input.get("timeout", self.default_timeout))
            driver = self._init_browser(headless)

            current_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            interactive_elements = "None"
            tabs = "None"
            prompt = get_system_prompt(
                user_input=input.get("query", ""),
                current_url=driver.current_url if driver else "",
                current_date=current_date,
                interactive_elements=interactive_elements,
                tabs=tabs
            )
            response = self.llm.generate(prompt=prompt)
            plan = self._parse_plan(response)

            if not plan.tasks:
                plan = BrowserPlan(tasks=[{"action": "done", "ui_index": None}])

            action_map = {
                "navigate": self._handle_navigation,
                "click": self._handle_click,
                "type": self._handle_typing,
                "wait": self._handle_wait,
                "wait_for_ajax": self._handle_wait_for_ajax,
                "scroll": self._handle_scroll,
                "hover": self._handle_hover,
                "screenshot": self._handle_screenshot,
                "switch_tab": self._handle_switch_tab,
                "execute_script": self._handle_execute_script,
                "drag_and_drop": self._handle_drag_and_drop,
                "select": self._handle_select_dropdown,
                "label_ui": self._handle_label_ui,
                "done": self._handle_done
            }

            task_index = 0
            overall_attempts = 0
            results = []
            current_url = driver.current_url if driver else ""

            while task_index < len(plan.tasks) and overall_attempts < self.max_overall_attempts:
                task = plan.tasks[task_index]
                handler = action_map.get(task.get("action", "").lower())
                if not handler:
                    results.append({"action": task.get("action"), "success": False, "message": "Action not supported"})
                    task_index += 1
                    continue

                result = self._execute_with_retries(driver, task, handler)
                results.append(result)
                if result.get("success", False):
                    task_index += 1
                else:
                    overall_attempts += 1

            return {"status": "success" if task_index == len(plan.tasks) else "partial", "results": results}

        except Exception as e:
            logger.exception("Execution error:")
            return {"status": "error", "message": str(e)}
        finally:
            if driver:
                driver.quit()

    def _init_browser(self, headless: bool) -> webdriver.Chrome:
        """Initialize browser with advanced security settings and performance optimizations."""
        options = Options()
        options.add_argument("--start-maximized")
        options.add_argument("--disable-blink-features=AutomationControlled")
        options.add_experimental_option("excludeSwitches", ["enable-automation"])
        options.add_experimental_option("useAutomationExtension", False)
        options.add_argument("--disable-dev-shm-usage")
        options.add_argument("--no-sandbox")
        if headless:
            options.add_argument("--headless=new")
        prefs = {
            "profile.managed_default_content_settings.images": 2,
            "profile.managed_default_content_settings.media_stream_mic": 2,
            "profile.managed_default_content_settings.media_stream_camera": 2
        }
        options.add_experimental_option('prefs', prefs)
        return webdriver.Chrome(
            service=Service(ChromeDriverManager().install()),
            options=options
        )

    def _parse_plan(self, response: str) -> BrowserPlan:
        """Robust JSON parsing with fallback ejection."""
        response = re.sub(r'//.*\n', '', response.strip())
        json_decoder = json.JSONDecoder()
        json_pos = response.find('{')
        if json_pos != -1:
            try:
                json_str, _ = json_decoder.raw_decode(response[json_pos:])
                plan = json.loads(json_str)
                return BrowserPlan.parse_obj({
                    "tasks": plan.get("action", [])
                })
            except json.JSONDecodeError:
                pass
        logger.error("Failed to parse valid JSON from LLM response")
        return BrowserPlan(tasks=[])

    def _execute_with_retries(self, driver, task, handler):
        """Retry logic with exponential backoff."""
        attempts = 0
        while attempts < self.max_retries:
            result = self._execute_safe_task(driver, task, handler)
            if result.get("success", False):
                return result
            attempts += 1
            time.sleep(2 ** attempts)  # Exponential backoff
        return {"action": task.get("action"), "success": False, "message": "Max retries exceeded"}

    def _execute_safe_task(self, driver: webdriver.Chrome, task: Dict[str, Any],
                             handler: Callable[[webdriver.Chrome, Dict[str, Any]], Dict[str, Any]]) -> Dict[str, Any]:
        """Error-safe task execution."""
        try:
            return handler(driver, task)
        except Exception as e:
            return {"action": task.get("action", "unknown"), "success": False, "message": f"Critical error: {str(e)}"}

    def _handle_label_ui(self, driver: webdriver.Chrome, task: Dict[str, Any]) -> Dict[str, Any]:
        """Enhanced UI labeling with explicit window-relative coordinates."""
        script = """
        (function() {
            const colors = {
                'INPUT': '#00FF00', 'BUTTON': '#0000FF', 'A': '#FFA500', 
                'TEXTAREA': '#32CD32', 'SELECT': '#FF69B4', 'DIV': '#FF4500'
            };
            const elements = document.querySelectorAll('*');
            const context = { indexes: [], overlays: [] };
            for (let i = 0; i < elements.length; i++) {
                const el = elements[i];
                const rect = el.getBoundingClientRect();
                const tagName = el.tagName.toUpperCase();
                if (rect.width > 10 && rect.height > 10 && Object.keys(colors).includes(tagName)) {
                    const overlay = document.createElement('div');
                    overlay.style.position = 'absolute';
                    overlay.style.border = `2px solid ${colors[tagName]}`;
                    overlay.style.left = `${rect.left + window.scrollX}px`;
                    overlay.style.top = `${rect.top + window.scrollY}px`;
                    overlay.style.width = `${rect.width}px`;
                    overlay.style.height = `${rect.height}px`;
                    overlay.style.zIndex = '99999';
                    overlay.style.pointerEvents = 'none';
                    
                    const label = document.createElement('div');
                    label.style.position = 'absolute';
                    label.style.backgroundColor = 'yellow';
                    label.style.color = 'black';
                    label.style.fontSize = '14px';
                    label.style.padding = '4px';
                    label.innerText = `${tagName} ${i + 1}`;
                    overlay.appendChild(label);
                    document.body.appendChild(overlay);
                    context.indexes.push({ index: i + 1, tagName, text: el.innerText.trim() });
                    context.overlays.push(overlay);
                }
            }
            return JSON.stringify(context);
        })();
        """
        try:
            result = json.loads(driver.execute_script(script))
            self.ui_mapping = result.get("indexes", [])
            return {"action": "label_ui", "success": True, "message": "UI elements labeled."}
        except Exception as e:
            logger.error(f"UI Labeling failed: {e}")
            return {"success": False, "message": "Unable to label UI elements"}

    def _handle_click(self, driver: webdriver.Chrome, task: Dict[str, Any]) -> Dict[str, Any]:
        """Reliable click handling with dynamic context awareness."""
        selector = task.get("selector", "")
        ui_index = task.get("ui_index")
        
        if ui_index is not None:
            elements = driver.find_elements(By.XPATH, f'//*[text() = "Element {ui_index}"]')
            if elements:
                target = elements[0]
                try:
                    target.click()
                    return {"success": True, "message": f"Clicked element {ui_index}"}
                except:
                    return {"success": False, "message": "Element not clickable"}

        try:
            element = WebDriverWait(driver, self.default_timeout).until(
                EC.element_to_be_clickable((By.CSS_SELECTOR, selector))
            )
            element.location_once_scrolled_into_view
            element.click()
            return {"success": True, "message": "Element clicked successfully"}
        except:
            return {"success": False, "message": "Element not found or not clickable"}

    def _handle_typing(self, driver: webdriver.Chrome, task: Dict[str, Any]) -> Dict[str, Any]:
        """Advanced text entry with fallback to fallback element detection."""
        selector = task.get("selector", "")
        text = task.get("value", "")
        try:
            element = WebDriverWait(driver, self.default_timeout).until(
                EC.presence_of_element_located((By.CSS_SELECTOR, selector))
            )
            element.send_keys(text)
            return {"success": True, "message": "Text entered successfully"}
        except:
            dynamic_selector = f'input[placeholder*="{text}"], textarea[placeholder*="{text}"]'
            element = WebDriverWait(driver, self.default_timeout).until(
                EC.presence_of_element_located((By.CSS_SELECTOR, dynamic_selector))
            )
            element.send_keys(text)
            return {"success": True, "message": "Fallback element found and text entered"}

    def _handle_navigation(self, driver: webdriver.Chrome, task: Dict[str, Any]) -> Dict[str, Any]:
        """Navigate to URL with explicit validation."""
        url = task.get("value", "")
        if not url.startswith(("http://", "https://")):
            url = f"https://{url}"
        try:
            driver.get(url)
            WebDriverWait(driver, self.default_timeout).until(
                EC.presence_of_element_located((By.TAG_NAME, "body"))
            )
            return {"success": True, "message": f"Navigated to {url}"}
        except:
            return {"success": False, "message": "Failed to navigate"}

    def _handle_wait(self, task: Dict[str, Any]) -> Dict[str, Any]:
        """Wait for a specified duration."""
        seconds = task.get("value", "0")
        try:
            time.sleep(float(seconds))
            return {"success": True, "message": f"Waited {seconds} seconds"}
        except:
            return {"success": False, "message": "Invalid wait time"}

    def _handle_wait_for_ajax(self, driver: webdriver.Chrome, task: Dict[str, Any]) -> Dict[str, Any]:
        """Wait for AJAX activity to settle."""
        try:
            WebDriverWait(driver, float(task.get("value", 15))).until(
                lambda d: d.execute_script("return (window.jQuery ? jQuery.active == 0 : true)")
            )
            return {"success": True, "message": "AJAX activity settled"}
        except:
            return {"success": False, "message": "AJAX wait failed"}

    def _handle_scroll(self, driver: webdriver.Chrome, task: Dict[str, Any]) -> Dict[str, Any]:
        """Scroll to element or bottom of page."""
        selector = task.get("selector", "")
        try:
            if selector:
                element = driver.find_element(By.CSS_SELECTOR, selector)
                driver.execute_script("arguments[0].scrollIntoView({behavior: 'smooth'});", element)
            else:
                driver.execute_script("window.scrollTo(0, document.body.scrollHeight);")
            return {"success": True, "message": "Scrolled successfully"}
        except:
            return {"success": False, "message": "Scroll failed"}

    def _handle_hover(self, driver: webdriver.Chrome, task: Dict[str, Any]) -> Dict[str, Any]:
        """Perform hover action on element."""
        selector = task.get("selector", "")
        try:
            element = driver.find_element(By.CSS_SELECTOR, selector)
            ActionChains(driver).move_to_element(element).perform()
            return {"success": True, "message": "Hovered successfully"}
        except:
            return {"success": False, "message": "Hover failed"}

    def _handle_screenshot(self, driver: webdriver.Chrome, task: Dict[str, Any]) -> Dict[str, Any]:
        """Capture and save screenshot."""
        filename = task.get("value", "screenshot.png")
        try:
            driver.save_screenshot(filename)
            return {"success": True, "message": f"Screenshot saved as {filename}"}
        except:
            return {"success": False, "message": "Failed to capture screenshot"}

    def _handle_switch_tab(self, driver: webdriver.Chrome, task: Dict[str, Any]) -> Dict[str, Any]:
        """Switch between browser tabs."""
        index = int(task.get("value", 0))
        try:
            tabs = driver.window_handles
            if 0 <= index < len(tabs):
                driver.switch_to.window(tabs[index])
                return {"success": True, "message": f"Switched to tab {index}"}
        except:
            pass
        return {"success": False, "message": "Failed to switch tab"}

    def _handle_execute_script(self, driver: webdriver.Chrome, task: Dict[str, Any]) -> Dict[str, Any]:
        """Execute arbitrary JavaScript."""
        script = task.get("value", "")
        try:
            driver.execute_script(script)
            return {"success": True, "message": "Script executed successfully"}
        except:
            return {"success": False, "message": "Script execution failed"}

    def _handle_drag_and_drop(self, driver: webdriver.Chrome, task: Dict[str, Any]) -> Dict[str, Any]:
        """Drag and drop elements."""
        source = driver.find_element(By.CSS_SELECTOR, task.get("source", ""))
        target = driver.find_element(By.CSS_SELECTOR, task.get("target", ""))
        try:
            ActionChains(driver).drag_and_drop(source, target).perform()
            return {"success": True, "message": "Drag and drop successful"}
        except:
            return {"success": False, "message": "Drag and drop failed"}

    def _handle_select_dropdown(self, driver: webdriver.Chrome, task: Dict[str, Any]) -> Dict[str, Any]:
        """Select option from dropdown."""
        selector = task.get("selector", "")
        option = task.get("value", "")
        try:
            element = Select(WebDriverWait(driver, self.default_timeout).until(
                EC.presence_of_element_located((By.CSS_SELECTOR, selector))
            ))
            element.select_by_visible_text(option)
            return {"success": True, "message": f"Selected '{option}' from dropdown"}
        except:
            return {"success": False, "message": "Dropdown selection failed"}

    def _handle_done(self, driver: webdriver.Chrome, task: Dict[str, Any]) -> Dict[str, Any]:
        """Mark task as complete."""
        return {"success": True, "message": "Task completed"}

    def _capture_screenshot_on_failure(self, driver: webdriver.Chrome, action: str):
        """Automated screenshot capture for failed actions."""
        filename = f"error_{action}_{datetime.now().strftime('%Y%m%d_%H%M%S')}.png"
        driver.save_screenshot(filename)
        logger.info(f"Failure screenshot saved as {filename}")